/* -*-C-*-

$Header: /usr/local/cvsrep/tbnl/specials.lisp,v 1.29 2004/11/09 14:10:57 edi Exp $

Copyright (c) 2003-2004, Chris Hanson <cph@csail.mit.edu> and Dr. Edmund Weitz <edi@agharta.de>.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS' AND ANY EXPRESSED OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

*/

#define VERSION_STRING "0.3"
#define READ_TIMEOUT 60000000

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "apr_date.h"
#include "apr_strings.h"
#include "apr_portable.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

module AP_MODULE_DECLARE_DATA lisp_module;

#define RELAY_ERROR(expr) do                                            \
{                                                                       \
  while (1)                                                             \
    {                                                                   \
      apr_status_t RELAY_ERROR_status = (expr);                         \
      if (APR_STATUS_IS_SUCCESS (RELAY_ERROR_status))                   \
        break;                                                          \
      if (!APR_STATUS_IS_EINTR (RELAY_ERROR_status))                    \
        return (RELAY_ERROR_status);                                    \
    }                                                                   \
} while (0)

#define ML_LOG_ERROR(status, r, msg)                                    \
  (ap_log_error (APLOG_MARK, APLOG_ERR, (status),                       \
                 ((r) -> server), (msg)))

#define ML_LOG_PERROR(r, msg)                                           \
  ML_LOG_ERROR ((APR_FROM_OS_ERROR (apr_get_os_error ())), (r), (msg))

#define RELAY_HTTP_ERROR(expr) do                                       \
{                                                                       \
  int RELAY_HTTP_ERROR_value = (expr);                                  \
  if (RELAY_HTTP_ERROR_value != OK)                                     \
    return (RELAY_HTTP_ERROR_value);                                    \
} while (0)

#define ENABLE_DEBUG 1
#ifdef ENABLE_DEBUG
#  define ML_LOG_DEBUG(r, ...)                                          \
    (ap_log_error (APLOG_MARK, APLOG_DEBUG, APR_SUCCESS,                        \
                   ((r) -> server), __VA_ARGS__))
#else
#  define ML_LOG_DEBUG(r, msg)
#endif

typedef struct lisp_cfg
{
  const char * server_address;
  apr_port_t server_port;
  const char * server_id;
  apr_uint64_t server_clearcount;
  unsigned int server_specified_p;
} lisp_cfg_t;

#define CFG(c) ((lisp_cfg_t *) (c))
#define CFG_REF(c, f) ((c) -> f)
#define SERVER_ADDRESS(c) CFG_REF (c, server_address)
#define SERVER_PORT(c) CFG_REF (c, server_port)
#define SERVER_ID(c) CFG_REF (c, server_id)
#define SERVER_CLEARCOUNT(c) CFG_REF (c, server_clearcount)
#define SERVER_SPECIFIED_P(c) CFG_REF (c, server_specified_p)

static apr_pool_t *ppool;
static apr_threadkey_t *tkey;

typedef struct lisp_private
{
  apr_pool_t *lisp_pool;
  apr_socket_t * lisp_socket;
  unsigned int
    lisp_socket_safe_p : 1,
    lisp_kill_pool_p : 1;
  char * lisp_server_address;
  long lisp_server_port;
  apr_uint64_t lisp_counter;
} lisp_private_t;

#define PRV(p) ((lisp_private_t *) (p))
#define PRV_REF(p, f) ((p) -> f)
#define LISP_POOL(p) PRV_REF (p, lisp_pool)
#define LISP_SOCKET(p) PRV_REF (p, lisp_socket)
#define LISP_SOCKET_SAFE_P(p) PRV_REF (p, lisp_socket_safe_p)
#define LISP_KILL_POOL_P(p) PRV_REF (p, lisp_kill_pool_p)
#define LISP_SERVER_ADDRESS(p) PRV_REF (p, lisp_server_address)
#define LISP_SERVER_PORT(p) PRV_REF (p, lisp_server_port)
#define LISP_COUNTER(p) PRV_REF (p, lisp_counter)

typedef struct lisp_conn {
  lisp_private_t *private_data;
} lisp_conn_t;

#define CONN_REF(c, f) ((c) -> f)
#define PRIVATE_DATA(c) CONN_REF (c, private_data)

typedef struct input_buffer
{
  char data [4096];
  const char * start;
  const char * end;
} input_buffer_t;

typedef const char * header_map_t (const char *);

/* Global variables */
static apr_pool_t * socket_pool = 0;

/* Forward references */
#ifdef ENABLE_DEBUG
static apr_status_t close_lisp_socket (lisp_conn_t * ccfg, request_rec * r);
#else
static apr_status_t close_lisp_socket (lisp_conn_t * ccfg);
#endif
static apr_status_t copy_headers
  (apr_table_t * table, header_map_t * map_name, apr_socket_t * socket);
static header_map_t map_env_var_to_lisp_header;
static header_map_t map_header_to_lisp_header;
static int write_client_data
  (request_rec * r, const char * data, unsigned int n_bytes);

/* Configuration data */

static lisp_cfg_t *
make_lisp_cfg (apr_pool_t * pool)
{
  return (apr_pcalloc (pool, (sizeof (lisp_cfg_t))));
}

static lisp_cfg_t *
default_lisp_cfg (apr_pool_t * pool)
{
  lisp_cfg_t * cfg = (make_lisp_cfg (pool));
  (SERVER_ADDRESS (cfg)) = (apr_pstrdup (pool, "127.0.0.1"));
  (SERVER_PORT (cfg)) = 3000;
  (SERVER_ID (cfg)) = (apr_pstrdup (pool, "apache"));
  (SERVER_SPECIFIED_P (cfg)) = 0;
  (SERVER_CLEARCOUNT (cfg)) = 1024;
  return (cfg);
}

static lisp_cfg_t *
copy_lisp_cfg (apr_pool_t * pool, lisp_cfg_t * cfg)
{
  lisp_cfg_t * copy = (make_lisp_cfg (pool));
  (SERVER_ADDRESS (copy)) = (SERVER_ADDRESS (cfg));
  (SERVER_PORT (copy)) = (SERVER_PORT (cfg));
  (SERVER_ID (copy)) = (SERVER_ID (cfg));
  (SERVER_SPECIFIED_P (copy)) = (SERVER_SPECIFIED_P (cfg));
  (SERVER_CLEARCOUNT (copy)) = (SERVER_CLEARCOUNT (cfg));
  return (copy);
}

static void *
lisp_create_dir_config (apr_pool_t * pool, char * directory)
{
  return (default_lisp_cfg (pool));
}

static void *
lisp_create_server_config (apr_pool_t * pool, server_rec * s)
{
  return (default_lisp_cfg (pool));
}

static void *
lisp_merge_config (apr_pool_t * pool, void * base_cfg, void * add_cfg)
{
  return
    ((SERVER_SPECIFIED_P (CFG (add_cfg)))
     ? (copy_lisp_cfg (pool, add_cfg))
     : (SERVER_SPECIFIED_P (CFG (base_cfg)))
     ? (copy_lisp_cfg (pool, base_cfg))
     : (default_lisp_cfg (pool)));
}

static const char *
lisp_set_clearcount (cmd_parms *cmd,
                     void * cfg_void,
                     const char *arg)
{
  lisp_cfg_t * cfg = cfg_void;
  const char *err;

  if ((err = ap_check_cmd_context (cmd, GLOBAL_ONLY)))
    return err;

  (SERVER_CLEARCOUNT (cfg)) = apr_atoi64(arg);

  return NULL;
}

static const char *
lisp_set_server (cmd_parms * cmd,
                 void * cfg_void,
                 const char * server_address,
                 const char * server_port,
                 const char * server_id)
{
  lisp_cfg_t * cfg = cfg_void;
  long port;
  {
    char * end;
    port = (strtol (server_port, (&end), 0));
    if (((*end) != '\0') || (port < 0))
      return ("malformed server port");
  }
  (SERVER_ADDRESS (cfg)) = (apr_pstrdup ((cmd->pool), server_address));
  (SERVER_PORT (cfg)) = port;
  (SERVER_ID (cfg)) = (apr_pstrdup ((cmd->pool), server_id));
  (SERVER_SPECIFIED_P (cfg)) = 1;
  return (0);
}

/* Socket to Lisp process */

static apr_status_t
#ifdef ENABLE_DEBUG
open_lisp_socket (lisp_cfg_t * cfg, lisp_conn_t * ccfg, request_rec * r)
#else
open_lisp_socket (lisp_cfg_t * cfg, lisp_conn_t * ccfg)
#endif
{
  apr_sockaddr_t * addr;
  apr_socket_t * socket;
  lisp_private_t * priv = (PRIVATE_DATA (ccfg));

  ML_LOG_DEBUG(r, "Thread %d got socket request for %s:%d", apr_os_thread_current(), (SERVER_ADDRESS (cfg)), (SERVER_PORT (cfg)));
  if ((LISP_SOCKET (priv)) != 0)
    {
      if ((LISP_SOCKET_SAFE_P (priv)) &&
          (!strcmp((SERVER_ADDRESS (cfg)), (LISP_SERVER_ADDRESS (priv)))) &&
          ((SERVER_PORT (cfg)) == (LISP_SERVER_PORT (priv))))
        {
          ML_LOG_DEBUG(r, "Thread %d re-uses socket %d", (LISP_SOCKET (priv)));
          return (APR_SUCCESS);
        }
      else
#ifdef ENABLE_DEBUG
        RELAY_ERROR (close_lisp_socket (ccfg, r));
#else
        RELAY_ERROR (close_lisp_socket (ccfg));
#endif
    }

  RELAY_ERROR
    (apr_sockaddr_info_get ((&addr), (SERVER_ADDRESS (cfg)), APR_UNSPEC,
                            (SERVER_PORT (cfg)), 0, socket_pool));

  RELAY_ERROR
    (apr_socket_create ((&socket), AF_INET, SOCK_STREAM, socket_pool));

  /* Should be apr_socket_connect */
  RELAY_ERROR (apr_connect (socket, addr));

  {
    input_buffer_t * buffer
      = (apr_palloc (socket_pool, (sizeof (input_buffer_t))));
    (buffer -> start) = (buffer -> data);
    (buffer -> end) = (buffer -> data);
    RELAY_ERROR (apr_socket_data_set (socket, buffer, "input-buffer", 0));
  }

  ML_LOG_DEBUG(r, "Thread %d created new socket %d", apr_os_thread_current(), socket);

  (LISP_SOCKET (priv)) = socket;
  (LISP_SOCKET_SAFE_P (priv)) = 0;
  (LISP_SERVER_ADDRESS (priv)) = (apr_pstrdup ((LISP_POOL (priv)), (SERVER_ADDRESS (cfg))));
  (LISP_SERVER_PORT (priv)) = (SERVER_PORT (cfg));
  return (APR_SUCCESS);
}

static apr_status_t
#ifdef ENABLE_DEBUG
close_lisp_socket (lisp_conn_t * ccfg, request_rec * r)
#else
close_lisp_socket (lisp_conn_t * ccfg)
#endif
{
  lisp_private_t * priv = (PRIVATE_DATA (ccfg));

  if (LISP_SOCKET (priv))
    {
      ML_LOG_DEBUG(r, "Thread %d closes socket %d", apr_os_thread_current(), (LISP_SOCKET (priv)));

      RELAY_ERROR (apr_socket_close (LISP_SOCKET (priv)));
      (LISP_SOCKET (priv)) = 0;
    }
  (LISP_SERVER_ADDRESS (priv)) = NULL;
  (LISP_SERVER_PORT (priv)) = 0;

  return (APR_SUCCESS);
}

static apr_status_t
write_lisp_data (apr_socket_t * socket,
                 const char * data, unsigned int n_bytes)
{
  const char * p = data;
  apr_size_t n1 = n_bytes;
  while (1)
    {
      apr_size_t n2 = n1;
      /* Should be apr_socket_send */
      RELAY_ERROR (apr_send (socket, p, (&n2)));
      if (n2 == n1)
        return (APR_SUCCESS);
      p += n2;
      n1 -= n2;
    }
}

static apr_status_t
write_lisp_line (apr_socket_t * socket, const char * data)
{
  RELAY_ERROR (write_lisp_data (socket, data, (strlen (data))));
  RELAY_ERROR (write_lisp_data (socket, "\n", 1));
  return (APR_SUCCESS);
}

static apr_status_t
write_lisp_header (apr_socket_t * socket,
                   const char * name, const char * value)
{
  RELAY_ERROR (write_lisp_line (socket, name));
  RELAY_ERROR (write_lisp_line (socket, value));
  return (APR_SUCCESS);
}

static apr_status_t
get_input_buffer (apr_socket_t * socket, input_buffer_t ** buffer_r)
{
  return (apr_socket_data_get (((void **) buffer_r), "input-buffer", socket));
}

static apr_status_t
fill_input_buffer (apr_socket_t * socket)
{
  input_buffer_t * buffer;
  apr_size_t length;

  RELAY_ERROR (get_input_buffer (socket, (&buffer)));
  RELAY_ERROR
    (((length = (sizeof (buffer->data))),
      (apr_recv (socket, (buffer->data), (&length)))));
  (buffer->start) = (buffer->data);
  (buffer->end) = ((buffer->data) + length);
  if (length == 0)
    (buffer->start) += 1;
  return (APR_SUCCESS);
}

static apr_status_t
read_lisp_line (apr_socket_t * socket, char * s, unsigned int len)
{
  input_buffer_t * buffer;
  char * scan_output = s;
  char * end_output = (scan_output + (len - 1));
  unsigned int n_pending_returns = 0;

  RELAY_ERROR (get_input_buffer (socket, (&buffer)));
  while (1)
    {
      if ((buffer->start) == (buffer->end))
        RELAY_ERROR (fill_input_buffer (socket));

      if ((buffer->start) > (buffer->end))
        {
          if (scan_output == s)
            return (APR_EOF);
          goto done;
        }

      while (((buffer->start) < (buffer->end)) && (scan_output < end_output))
        {
          char c = (*(buffer->start)++);
          if (c == '\n')
            {
              if (n_pending_returns > 0)
                n_pending_returns -= 1;
              goto done;
            }
          if (c == '\r')
            n_pending_returns += 1;
          else
            {
              while ((n_pending_returns > 0) && (scan_output < end_output))
                {
                  (*scan_output++) = '\r';
                  n_pending_returns -= 1;
                }
              if (scan_output == end_output)
                goto done;
              (*scan_output++) = c;
            }
        }
    }
 done:
  while ((n_pending_returns > 0) && (scan_output < end_output))
    {
      (*scan_output++) = '\r';
      n_pending_returns -= 1;
    }
  (*scan_output) = '\0';
  return (APR_SUCCESS);
}

/* Request handler */

#ifdef ENABLE_DEBUG
#define CVT_ERROR(expr, msg) do                                         \
{                                                                       \
  apr_status_t CVT_ERROR_status = (expr);                               \
  if (!APR_STATUS_IS_SUCCESS (CVT_ERROR_status))                        \
    {                                                                   \
      ap_log_error (APLOG_MARK, APLOG_ERR, CVT_ERROR_status,            \
                    (r->server), "error %s", msg);                      \
      close_lisp_socket (ccfg, r);                                      \
      return (HTTP_INTERNAL_SERVER_ERROR);                              \
    }                                                                   \
} while (0)
#else
#define CVT_ERROR(expr, msg) do                                         \
{                                                                       \
  apr_status_t CVT_ERROR_status = (expr);                               \
  if (!APR_STATUS_IS_SUCCESS (CVT_ERROR_status))                        \
    {                                                                   \
      ap_log_error (APLOG_MARK, APLOG_ERR, CVT_ERROR_status,            \
                    (r->server), "error %s", msg);                      \
      close_lisp_socket (ccfg);                                         \
      return (HTTP_INTERNAL_SERVER_ERROR);                              \
    }                                                                   \
} while (0)
#endif

static int
lisp_handler (request_rec * r)
{
  lisp_cfg_t * cfg
    = (ap_get_module_config ((r->per_dir_config), (&lisp_module)));
  conn_rec *c = r->connection;
  lisp_conn_t * ccfg
    = (ap_get_module_config ((c->conn_config), (&lisp_module)));
  lisp_private_t * priv;
  int content_length = (-1);
  int keep_socket_p = 0;
  apr_socket_t * socket;

  if ((strcmp ((r->handler), "lisp-handler")) != 0)
    return (DECLINED);

  if (!(cfg && ccfg))
    {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, APR_EGENERAL, r,
                    "mod_lisp2: missing configuration data");
      return HTTP_INTERNAL_SERVER_ERROR;
    }

  priv = (PRIVATE_DATA (ccfg));

  if ((SERVER_CLEARCOUNT (cfg)) > 0 && (LISP_COUNTER (priv))++ > (SERVER_CLEARCOUNT (cfg)))
    (LISP_KILL_POOL_P (priv)) = 1;

  /* Open a connection to the Lisp process.  */
  ML_LOG_DEBUG (r, "open lisp connection");
#ifdef ENABLE_DEBUG
  CVT_ERROR ((open_lisp_socket (cfg, ccfg, r)), "opening connection to Lisp");
#else
  CVT_ERROR ((open_lisp_socket (cfg, ccfg)), "opening connection to Lisp");
#endif
  (LISP_SOCKET_SAFE_P (priv)) = 0;
  socket = (LISP_SOCKET (priv));

  /* Remove any timeout that might be left over from earlier.  */
  ML_LOG_DEBUG (r, "clear socket timeout");
  CVT_ERROR ((apr_socket_timeout_set (socket, (-1))), "clearing read timeout");
  
  /* Convert environment variables to headers and send them.  */
  ML_LOG_DEBUG (r, "write env-var headers");
  ap_add_cgi_vars (r);
  ap_add_common_vars (r);
  if ((r->subprocess_env) != 0)
    CVT_ERROR
      ((copy_headers
        ((r->subprocess_env), map_env_var_to_lisp_header, socket)),
       "writing to Lisp");

  /* Send this before client headers so ASSOC can be used to grab it
     without worrying about some joker sending a server-id header of
     his own.  (Robert Macomber) */
  ML_LOG_DEBUG (r, "write headers");
  CVT_ERROR ((write_lisp_header (socket, "server-id", (cfg->server_id))),
             "writing to Lisp");

  CVT_ERROR ((write_lisp_header (socket, "server-baseversion", AP_SERVER_BASEVERSION)),
             "writing to Lisp");
  CVT_ERROR ((write_lisp_header (socket, "modlisp-version", VERSION_STRING)),
             "writing to Lisp");
  CVT_ERROR ((write_lisp_header (socket, "modlisp-major-version", "2")),
             "writing to Lisp");
  /* Send all the remaining headers.  */
  if ((r->headers_in) != 0)
    CVT_ERROR
      ((copy_headers ((r->headers_in), map_header_to_lisp_header, socket)),
       "writing to Lisp");

  /* Send the end-of-headers marker.  */
  ML_LOG_DEBUG (r, "write end-of-headers");
  CVT_ERROR ((write_lisp_line (socket, "end")), "writing to Lisp");

  /* Send the request entity.  */
  RELAY_HTTP_ERROR (ap_setup_client_block (r, REQUEST_CHUNKED_DECHUNK));
  if (ap_should_client_block (r))
    {
      char buffer [4096];
      ML_LOG_DEBUG (r, "write entity");
      while (1)
        {
          long n_read = (ap_get_client_block (r, buffer, (sizeof (buffer))));
          if (n_read < 0)
            {
              ML_LOG_PERROR (r, "error reading from client");
#ifdef ENABLE_DEBUG
              close_lisp_socket (ccfg, r);
#else
              close_lisp_socket (ccfg);
#endif
              return (HTTP_INTERNAL_SERVER_ERROR);
            }
          if (n_read == 0)
            break;

          {
            apr_status_t status = (write_lisp_data (socket, buffer, n_read));
            if (!APR_STATUS_IS_SUCCESS (status))
              {
                while ((ap_get_client_block (r, buffer, (sizeof (buffer))))
                       > 0)
                  ;
                ML_LOG_ERROR (status, r, "writing to Lisp");
#ifdef ENABLE_DEBUG
                close_lisp_socket (ccfg, r);
#else
                close_lisp_socket (ccfg);
#endif
                return (HTTP_INTERNAL_SERVER_ERROR);
              }
          }
        }
    }

  /* Set up read timeout so we don't hang forever if Lisp is wedged.  */
  ML_LOG_DEBUG (r, "set socket timeout");
  CVT_ERROR ((apr_socket_timeout_set (socket, READ_TIMEOUT)),
             "setting read timeout");

  /* Read the headers and process them.  */
  ML_LOG_DEBUG (r, "read headers");
  while (1)
    {
      char header_name [4096];
      char header_value [MAX_STRING_LEN];
      CVT_ERROR
        ((read_lisp_line (socket, header_name, (sizeof (header_name)))),
         "reading from Lisp");

      if ((strcasecmp (header_name, "end")) == 0)
        break;

      CVT_ERROR
        ((read_lisp_line (socket, header_value, (sizeof (header_value)))),
         "reading from Lisp");

      if ((strcasecmp (header_name, "content-type")) == 0)
        {
          char * tmp = (apr_pstrdup ((r->pool), header_value));
          ap_content_type_tolower (tmp);
          (r->content_type) = tmp;
        }
      else if ((strcasecmp (header_name, "status")) == 0)
        {
          (r->status) = (atoi (header_value));
          (r->status_line) = (apr_pstrdup ((r->pool), header_value));
        }
      else if ((strcasecmp (header_name, "location")) == 0)
        apr_table_set ((r->headers_out), header_name, header_value);
      else if ((strcasecmp (header_name, "content-length")) == 0)
        {
          apr_table_set ((r->headers_out), header_name, header_value);
          content_length = (atoi (header_value));
        }
      else if ((strcasecmp (header_name, "last-modified")) == 0)
        {
          time_t mtime = (apr_date_parse_http (header_value));
          ap_update_mtime (r, mtime);
          ap_set_last_modified (r);
        }
      else if ((strcasecmp (header_name, "keep-socket")) == 0)
        keep_socket_p = (atoi (header_value));
      else if ((strcasecmp (header_name, "log-emerg")) == 0)
        ap_log_error (APLOG_MARK, APLOG_EMERG, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-alert")) == 0)
        ap_log_error (APLOG_MARK, APLOG_ALERT, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-crit")) == 0)
        ap_log_error (APLOG_MARK, APLOG_CRIT, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-error")) == 0)
        ap_log_error (APLOG_MARK, APLOG_ERR, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-warning")) == 0)
        ap_log_error (APLOG_MARK, APLOG_WARNING, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-notice")) == 0)
        ap_log_error (APLOG_MARK, APLOG_NOTICE, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-info")) == 0)
        ap_log_error (APLOG_MARK, APLOG_INFO, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log-debug")) == 0)
        ap_log_error (APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "log")) == 0)
        ap_log_error (APLOG_MARK, APLOG_ERR, APR_SUCCESS, (r->server),
                      "%s", header_value);
      else if ((strcasecmp (header_name, "note")) == 0)
        {
          char * p = (strchr (header_value, ' '));
          if (p != 0)
            {
              (*p++) = '\0';
              apr_table_setn ((r->notes),
                              (apr_pstrdup ((r->pool), header_value)),
                              (apr_pstrdup ((r->pool), p)));
            }
        }
      else if ((strcasecmp (header_name, "set-cookie")) == 0)
        {
          apr_table_add ((r->headers_out), header_name, header_value);
        }
      else
        apr_table_set ((r->headers_out), header_name, header_value);
    }

  /* Copy the reply entity from Lisp to the client.  */
  if (!r->header_only)
    {
      unsigned int n_read = 0;
      input_buffer_t * buffer;

      ML_LOG_DEBUG (r, "read entity");
      CVT_ERROR ((get_input_buffer (socket, (&buffer))), "reading from Lisp");
      while ((buffer->start) <= (buffer->end))
        {
          unsigned int n_bytes = ((buffer->end) - (buffer->start));
          n_read += n_bytes;
          if ((content_length >= 0) && (n_read > content_length))
            {
              n_bytes -= (n_read - content_length);
              n_read -= (n_read - content_length);
            }
          if (!write_client_data (r, (buffer->start), n_bytes))
            {
#ifdef ENABLE_DEBUG
              close_lisp_socket (ccfg, r);
#else
              close_lisp_socket (ccfg);
#endif
              return (HTTP_INTERNAL_SERVER_ERROR);
            }
          (buffer->start) += n_bytes;
          if (n_read == content_length)
            break;
          CVT_ERROR ((fill_input_buffer (socket)), "reading from Lisp");
        }
    }
  if ((content_length < 0) || (!keep_socket_p))
#ifdef ENABLE_DEBUG
    CVT_ERROR ((close_lisp_socket (ccfg, r)), "closing connection to Lisp");
#else
    CVT_ERROR ((close_lisp_socket (ccfg)), "closing connection to Lisp");
#endif
  else
    (LISP_SOCKET_SAFE_P (priv)) = 1;
  ML_LOG_DEBUG (r, "request finished");
  return (OK);
}

static apr_status_t
copy_headers (apr_table_t * table, header_map_t * map_name,
              apr_socket_t * socket)
{
  const apr_array_header_t * h = (apr_table_elts (table));
  const apr_table_entry_t * scan = ((apr_table_entry_t *) (h->elts));
  const apr_table_entry_t * end = (scan + (h->nelts));
  while (scan < end)
    {
      const char * name = ((*map_name) (scan->key));
      if (name != 0)
        RELAY_ERROR (write_lisp_header (socket, name, (scan->val)));
      scan += 1;
    }
  return (APR_SUCCESS);
}

static const char *
map_env_var_to_lisp_header (const char * var)
{
  const char * plist [] = {
    "REQUEST_URI", "url",
    "CONTENT_TYPE", "content-type",
    "CONTENT_LENGTH", "content-length",
    "REQUEST_METHOD", "method",
    "REMOTE_ADDR", "remote-ip-addr",
    "REMOTE_PORT", "remote-ip-port",
    "SERVER_ADDR", "server-ip-addr",
    "SERVER_PORT", "server-ip-port",
    "SERVER_PROTOCOL", "server-protocol",
    "SCRIPT_FILENAME", "script-filename",
    "SSL_SESSION_ID", "ssl-session-id",
    0
    };
  const char ** p = plist;
  if ((var == 0) || ((strncmp (var, "HTTP_", 5)) == 0))
    return (0);
  while (1)
    {
      const char * v = (*p++);
      if (v == 0)
        return (0);
      if ((strcmp (v, var)) == 0)
        return (*p);
      p += 1;
    }
}

static const char *
map_header_to_lisp_header (const char * name)
{
  return
    (((name != 0) && ((strcasecmp (name, "end")) == 0))
     ? "end-header"
     : name);
}

static int
write_client_data (request_rec * r, const char * data, unsigned int n_bytes)
{
  while (n_bytes > 0)
    {
      int n = (ap_rwrite (data, n_bytes, r));
      if (n < 0)
        {
          ML_LOG_PERROR (r, "error writing to client");
          return (0);
        }
      n_bytes -= n;
      data += n;
    }
  return (1);
}

/* Module setup */

static int
lisp_post_config (apr_pool_t * cfg_pool,
                  apr_pool_t * log_pool,
                  apr_pool_t * temp_pool,
                  server_rec * s)
{
  ap_add_version_component (cfg_pool, "mod_lisp2/" VERSION_STRING);
  return (OK);
}

static apr_status_t
destroy_socket_pool (void * dummy)
{
  apr_pool_destroy (socket_pool);
  socket_pool = 0;
  return (APR_SUCCESS);
}

static void
private_destroy (void *data)
{
  apr_pool_destroy((LISP_POOL (PRV (data))));
}

static void
lisp_child_init (apr_pool_t * pool, server_rec * s)
{
  apr_status_t status;

  ppool = s->process->pool;

  if ((status = apr_threadkey_private_create(&tkey, private_destroy, ppool)) != APR_SUCCESS)
    ap_log_error (APLOG_MARK, APLOG_CRIT, status, s, "mod_lisp2: cannot create thread private key");

  if (APR_STATUS_IS_SUCCESS (apr_pool_create ((&socket_pool), 0)))
    apr_pool_cleanup_register
      (pool, 0, destroy_socket_pool, destroy_socket_pool);
}

lisp_private_t *
create_private_data (apr_pool_t *pool)
{
  lisp_private_t *priv;
  apr_pool_t *tpool;

  if (apr_pool_create (&tpool, pool) != APR_SUCCESS)
    return NULL;

  if (!(priv = apr_pcalloc (tpool, sizeof (*priv))))
    {
      apr_pool_destroy(tpool);
      return NULL;
    }

  (LISP_POOL (priv)) = tpool;

  return priv;
}

static int
lisp_pre_conn (conn_rec *c, void *csd)
{
  lisp_conn_t *ccfg = apr_pcalloc (c->pool, sizeof (*ccfg));
  apr_status_t status;

  if (!ccfg)
    {
      ap_log_error (APLOG_MARK, APLOG_CRIT, APR_ENOMEM, c->base_server,
                    "mod_lisp2: cannot create connection configuration data");
      return HTTP_INTERNAL_SERVER_ERROR;
    }

  if ((status = apr_threadkey_private_get ((void *) (&(PRIVATE_DATA (ccfg))), tkey)) != APR_SUCCESS)
    {
      ap_log_error (APLOG_MARK, APLOG_CRIT, status, c->base_server,
                    "mod_lisp2: cannot get thread private data");
      return HTTP_INTERNAL_SERVER_ERROR;
    }

  if ((PRIVATE_DATA (ccfg)) && (LISP_KILL_POOL_P (PRIVATE_DATA (ccfg)))) {
    apr_pool_destroy ((LISP_POOL (PRIVATE_DATA (ccfg))));
    (PRIVATE_DATA (ccfg)) = NULL;
  }

  if (!(PRIVATE_DATA (ccfg))) {
    if (!((PRIVATE_DATA (ccfg)) = create_private_data (ppool)))
      {
        ap_log_error (APLOG_MARK, APLOG_CRIT, APR_ENOMEM, c->base_server,
                      "mod_lisp2: cannot create thread private data");
        return HTTP_INTERNAL_SERVER_ERROR;
      }
    if (apr_threadkey_private_set((PRIVATE_DATA (ccfg)), tkey) != APR_SUCCESS)
      {
        apr_pool_destroy ((LISP_POOL (PRIVATE_DATA (ccfg))));
        ap_log_error (APLOG_MARK, APLOG_CRIT, status, c->base_server,
                      "mod_lisp2: cannot set thread private data");
        return HTTP_INTERNAL_SERVER_ERROR;
      }
  }

  ap_set_module_config(c->conn_config, &lisp_module, ccfg);

  return OK;
}

static const command_rec lisp_command_handlers [] =
{
  AP_INIT_TAKE3
    ("LispServer", lisp_set_server, 0, OR_ALL,
     "The Lisp server name, port and ID string."
     "  Example: LispServer 127.0.0.1 3000 \"apache\""),
  AP_INIT_TAKE1
    ("LispClearCount", lisp_set_clearcount, 0, RSRC_CONF,
     "After how many requests the private pool gets cleared"
     "  Example: LispClearCount 1024"),
  {0}
};

static void
register_hooks (apr_pool_t * pool)
{
  ap_hook_post_config (lisp_post_config, 0, 0, APR_HOOK_MIDDLE);
  ap_hook_child_init (lisp_child_init, 0, 0, APR_HOOK_MIDDLE);
  ap_hook_pre_connection (lisp_pre_conn, 0, 0, APR_HOOK_MIDDLE);
  ap_hook_handler (lisp_handler, 0, 0, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA lisp_module =
{
  STANDARD20_MODULE_STUFF,
  lisp_create_dir_config,       /* create per-directory config structures */
  lisp_merge_config,            /* merge per-directory config structures */
  lisp_create_server_config,    /* create per-server config structures */
  lisp_merge_config,            /* merge per-server config structures */
  lisp_command_handlers,
  register_hooks
};
